/*--------------------------------------------------------
  "THE BEER-WARE LICENSE" (Revision 42):
  Alex Kostyuk wrote this code. As long as you retain this
  notice you can do whatever you want with this stuff.
  If we meet some day, and you think this stuff is worth it,
  you can buy me a beer in return.
----------------------------------------------------------*/

#include "includes.h"

CONFIG config;
static int cfg_item = -1;
static int ind_item = -1;
static int port = 0;

int delay_after_shift = 20;
int pressing_duration = 60;


typedef struct
{
    const char *group;
    const char *key;
    char const *var;
    int size;
} INIT_STRUCT;



KEY_CONFIG key_config[MAX_KEYS];

int key_parser(char *outp, char *inp);

int key_conversion(int ev, int port)
{
    int ev_type = (ev << 4) & 0xF000;
    ev &= 0xFF;
    ev |= (ev_type | (port << 8));
    return ev;
}


int find_key(const DICT *dct, char *key)
{
    int i;

    for(i=0; dct[i].key[0]; i++)
    {
        if(0 == strcmp(dct[i].key,key))
            return dct[i].val;
    }

    return -1;
}

#define ADD_ITEM(x,y,z)  {x,y,config.z,sizeof(config.z)}

static const INIT_STRUCT init_struct[] =
{

    ADD_ITEM("mode",   "log",                  mode_log),
    ADD_ITEM("mode",   "console_log",          mode_console_log),
    ADD_ITEM("app",    "app_name",             app_capbar_text),

    {NULL,NULL,NULL,-1}
};

static int config_cb(int lineno, void* user, const char* section, const char* name,
                     const char* value)
{
    int i;
    int key_id,mode,type;
    char pattern[100];
#define MATCH(s, n) strcmp(section, s) == 0 && strcmp(name, n) == 0

    for(i=0; init_struct[i].size!=-1; i++)
    {
        if(MATCH(init_struct[i].group, init_struct[i].key))
        {
            strncpy((char*)init_struct[i].var,value,init_struct[i].size);
        }
    }

    if(MATCH("device","port"))
    {
        port = atoi(value);
    }


    if(MATCH("app",    "delay_after_shift"))
    {
        delay_after_shift = atoi(value);
    }

    if(MATCH("app",    "pressing_duration"))
    {
        pressing_duration = atoi(value);
    }




    for(i=0; i<SERIAL_PORTS_MAX; i++)
    {
        sprintf(pattern,"port_%d",i);
        if(strcmp(section, "serial") == 0 && strcmp(name, pattern) == 0)
        {
            strcpy(sws[i].serial_port,value);
        }
    }

    if(
        (   section[0] == 'K'
            && section[1] == 'e'
            && section[2] == 'y') )
    {
        sscanf(&section[3],"%d",&key_id);
        //key_id = key_conversion(key_id, port);

        /*

        #   254
        [Key08]
        on_press: APPS,KEY_08
        on_release:
        delay_after_shift: 5
        pressing_duration: 40

        */

        if(0 == strcmp(name, "on"))
        {
            key_config[key_id + port * 256].port = port;
            key_parser(key_config[key_id + port * 256].on,value);

        }
        else if(0 == strcmp(name, "off"))
        {
            key_config[key_id + port * 256].port = port;
            key_parser(key_config[key_id + port * 256].off,value);
        }

        else if(0 == strcmp(name, "hold"))
        {
            key_config[key_id + port * 256].port = port;
            key_parser(key_config[key_id + port * 256].hold,value);

        }
        else if(0 == strcmp(name, "release"))
        {
            key_config[key_id + port * 256].port = port;
            key_parser(key_config[key_id + port * 256].release,value);
        }
        else if(0 == strcmp(name, "delay_after_shift"))
        {
            key_config[key_id + port * 256].delay_after_shift = atoi(value);
        }
        else if(0 == strcmp(name, "pressing_duration"))
        {
            key_config[key_id + port * 256].pressing_duration = atoi(value);
        }

    }

    return 1;
}


int load_config(char *path)
{
    memset(key_config,0,sizeof(key_config));
    cfg_item = -1;
    ind_item = -1;
    port = 0;

    char ini_path[MAX_PATH];
    strcpy(ini_path,path);

    strcat(ini_path, MAPPER_INI);

    if(ini_parse(ini_path, config_cb, &config) < 0)
    {
        log_write("could not load config file %s",ini_path);
        return 0;
    }

    return 1;
}

int key_parser(char *outp, char *inp)
{
    int i,j,cnt,tok;
    char token[100];

    for(j=0,i=0,cnt=0;; i++)
    {

        if (inp[i] == 0x00)
        {
            /* end of line */
            token[j] = 0x00;
            j = 0;
            tok = key_find(token);
            if(tok != -1)
            {
                outp[cnt++] = tok;
            }
            outp[cnt] = 0x00;
            return cnt;

        }
        else if(strchr(",;.:-+",inp[i]))
        {
            /*separator*/
            token[j] = 0x00;
            j = 0;
            tok = key_find(token);
            if(tok != -1)
            {
                outp[cnt++] = tok;
            }

        }
        else if (strchr(" \t\r\n",inp[i]))
        {
            /* whitespace */

        }

        else
        {
            /* just a symbol */
            token[j++] = inp[i];
        }
    }


    return cnt;
}


