/*--------------------------------------------------------
  "THE BEER-WARE LICENSE" (Revision 42):
  Alex Kostyuk wrote this code. As long as you retain this
  notice you can do whatever you want with this stuff.
  If we meet some day, and you think this stuff is worth it,
  you can buy me a beer in return.
----------------------------------------------------------*/


#define WINVER 0x500
#include "includes.h"
#include <windows.h>
#include <wincon.h>
#include <conio.h>


#define BUTTON_RELEASED 0x400
#define ROTARY_SWITCH_ON_POSITION 0x100
#define BUTTON_PRESSED  0x800

static HANDLE key_mutex = INVALID_HANDLE_VALUE;

typedef struct
{
    const char* key;
    int value;
} KEY_TABLE;


KEY_TABLE key_table[] =
{
    {"ABNT_C1",VK_ABNT_C1},
    {"ABNT_C2",VK_ABNT_C2},
    {"ADD",VK_ADD},
    {"ATTN",VK_ATTN},
    {"BACK",VK_BACK},
    {"CANCEL",VK_CANCEL},
    {"CLEAR",VK_CLEAR},
    {"CRSEL",VK_CRSEL},
    {"DECIMAL",VK_DECIMAL},
    {"DIVIDE",VK_DIVIDE},
    {"EREOF",VK_EREOF},
    {"ESCAPE",VK_ESCAPE},
    {"EXECUTE",VK_EXECUTE},
    {"EXSEL",VK_EXSEL},
    {"ICO_CLEAR",VK_ICO_CLEAR},
    {"ICO_HELP",VK_ICO_HELP},
    {"KEY_0",VK_KEY_0},
    {"KEY_1",VK_KEY_1},
    {"KEY_2",VK_KEY_2},
    {"KEY_3",VK_KEY_3},
    {"KEY_4",VK_KEY_4},
    {"KEY_5",VK_KEY_5},
    {"KEY_6",VK_KEY_6},
    {"KEY_7",VK_KEY_7},
    {"KEY_8",VK_KEY_8},
    {"KEY_9",VK_KEY_9},
    {"KEY_A",VK_KEY_A},
    {"KEY_B",VK_KEY_B},
    {"KEY_C",VK_KEY_C},
    {"KEY_D",VK_KEY_D},
    {"KEY_E",VK_KEY_E},
    {"KEY_F",VK_KEY_F},
    {"KEY_G",VK_KEY_G},
    {"KEY_H",VK_KEY_H},
    {"KEY_I",VK_KEY_I},
    {"KEY_J",VK_KEY_J},
    {"KEY_K",VK_KEY_K},
    {"KEY_L",VK_KEY_L},
    {"KEY_M",VK_KEY_M},
    {"KEY_N",VK_KEY_N},
    {"KEY_O",VK_KEY_O},
    {"KEY_P",VK_KEY_P},
    {"KEY_Q",VK_KEY_Q},
    {"KEY_R",VK_KEY_R},
    {"KEY_S",VK_KEY_S},
    {"KEY_T",VK_KEY_T},
    {"KEY_U",VK_KEY_U},
    {"KEY_V",VK_KEY_V},
    {"KEY_W",VK_KEY_W},
    {"KEY_X",VK_KEY_X},
    {"KEY_Y",VK_KEY_Y},
    {"KEY_Z",VK_KEY_Z},
    {"MULTIPLY",VK_MULTIPLY},
    {"NONAME",VK_NONAME},
    {"NUMPAD0",VK_NUMPAD0},
    {"NUMPAD1",VK_NUMPAD1},
    {"NUMPAD2",VK_NUMPAD2},
    {"NUMPAD3",VK_NUMPAD3},
    {"NUMPAD4",VK_NUMPAD4},
    {"NUMPAD5",VK_NUMPAD5},
    {"NUMPAD6",VK_NUMPAD6},
    {"NUMPAD7",VK_NUMPAD7},
    {"NUMPAD8",VK_NUMPAD8},
    {"NUMPAD9",VK_NUMPAD9},
    {"OEM_1",VK_OEM_1},
    {"OEM_102",VK_OEM_102},    {"OEM_2",VK_OEM_2},
    {"OEM_3",VK_OEM_3},
    {"OEM_4",VK_OEM_4},
    {"OEM_5",VK_OEM_5},
    {"OEM_6",VK_OEM_6},
    {"OEM_7",VK_OEM_7},
    {"OEM_8",VK_OEM_8},
    {"OEM_ATTN",VK_OEM_ATTN},
    {"OEM_AUTO",VK_OEM_AUTO},
    {"OEM_AX",VK_OEM_AX},
    {"OEM_BACKTAB",VK_OEM_BACKTAB},
    {"OEM_CLEAR",VK_OEM_CLEAR},
    {"OEM_COMMA",VK_OEM_COMMA},
    {"OEM_COPY",VK_OEM_COPY},
    {"OEM_CUSEL",VK_OEM_CUSEL},
    {"OEM_ENLW",VK_OEM_ENLW},
    {"OEM_FINISH",VK_OEM_FINISH},
    {"OEM_FJ_LOYA",VK_OEM_FJ_LOYA},
    {"OEM_FJ_MASSHOU",VK_OEM_FJ_MASSHOU},
    {"OEM_FJ_ROYA",VK_OEM_FJ_ROYA},
    {"OEM_FJ_TOUROKU",VK_OEM_FJ_TOUROKU},
    {"OEM_JUMP",VK_OEM_JUMP},
    {"OEM_MINUS",VK_OEM_MINUS},
    {"OEM_PA1",VK_OEM_PA1},
    {"OEM_PA2",VK_OEM_PA2},
    {"OEM_PA3",VK_OEM_PA3},
    {"OEM_PERIOD",VK_OEM_PERIOD},
    {"OEM_PLUS",VK_OEM_PLUS},
    {"OEM_RESET",VK_OEM_RESET},
    {"OEM_WSCTRL",VK_OEM_WSCTRL},
    {"PA1",VK_PA1},
    {"PACKET",VK_PACKET},
    {"PLAY",VK_PLAY},
    {"PROCESSKEY",VK_PROCESSKEY},
    {"RETURN",VK_RETURN},
    {"SELECT",VK_SELECT},
    {"SEPARATOR",VK_SEPARATOR},
    {"SPACE",VK_SPACE},
    {"SUBTRACT",VK_SUBTRACT},
    {"TAB",VK_TAB},
    {"ZOOM",VK_ZOOM},
    {"_none_",VK__none_},
    {"ACCEPT",VK_ACCEPT},
    {"APPS",VK_APPS},
    {"BROWSER_BACK",VK_BROWSER_BACK},
    {"BROWSER_FAVORITES",VK_BROWSER_FAVORITES},
    {"BROWSER_FORWARD",VK_BROWSER_FORWARD},
    {"BROWSER_HOME",VK_BROWSER_HOME},
    {"BROWSER_REFRESH",VK_BROWSER_REFRESH},
    {"BROWSER_SEARCH",VK_BROWSER_SEARCH},
    {"BROWSER_STOP",VK_BROWSER_STOP},
    {"CAPITAL",VK_CAPITAL},
    {"CONVERT",VK_CONVERT},
    {"DELETE",VK_DELETE},
    {"DOWN",VK_DOWN},
    {"END",VK_END},
    {"F1",VK_F1},
    {"F10",VK_F10},
    {"F11",VK_F11},
    {"F12",VK_F12},
    {"F13",VK_F13},
    {"F14",VK_F14},
    {"F15",VK_F15},
    {"F16",VK_F16},
    {"F17",VK_F17},
    {"F18",VK_F18},
    {"F19",VK_F19},
    {"F2",VK_F2},
    {"F20",VK_F20},
    {"F21",VK_F21},
    {"F22",VK_F22},
    {"F23",VK_F23},
    {"F24",VK_F24},
    {"F3",VK_F3},
    {"F4",VK_F4},
    {"F5",VK_F5},
    {"F6",VK_F6},
    {"F7",VK_F7},
    {"F8",VK_F8},
    {"F9",VK_F9},
    {"FINAL",VK_FINAL},
    {"HELP",VK_HELP},
    {"HOME",VK_HOME},
    {"ICO_00",VK_ICO_00},
    {"INSERT",VK_INSERT},
    {"JUNJA",VK_JUNJA},
    {"KANA",VK_KANA},
    {"KANJI",VK_KANJI},
    {"LAUNCH_APP1",VK_LAUNCH_APP1},
    {"LAUNCH_APP2",VK_LAUNCH_APP2},
    {"LAUNCH_MAIL",VK_LAUNCH_MAIL},
    {"LAUNCH_MEDIA_SELECT",VK_LAUNCH_MEDIA_SELECT},
    {"LBUTTON",VK_LBUTTON},
    {"LCONTROL",VK_LCONTROL},
    {"LEFT",VK_LEFT},
    {"LMENU",VK_LMENU},
    {"LSHIFT",VK_LSHIFT},
    {"LWIN",VK_LWIN},
    {"MBUTTON",VK_MBUTTON},
    {"MEDIA_NEXT_TRACK",VK_MEDIA_NEXT_TRACK},
    {"MEDIA_PLAY_PAUSE",VK_MEDIA_PLAY_PAUSE},
    {"MEDIA_PREV_TRACK",VK_MEDIA_PREV_TRACK},
    {"MEDIA_STOP",VK_MEDIA_STOP},
    {"MODECHANGE",VK_MODECHANGE},
    {"NEXT",VK_NEXT},
    {"NONCONVERT",VK_NONCONVERT},
    {"NUMLOCK",VK_NUMLOCK},
    {"OEM_FJ_JISHO",VK_OEM_FJ_JISHO},
    {"PAUSE",VK_PAUSE},
    {"PRINT",VK_PRINT},
    {"PRIOR",VK_PRIOR},
    {"RBUTTON",VK_RBUTTON},
    {"RCONTROL",VK_RCONTROL},
    {"RIGHT",VK_RIGHT},
    {"RMENU",VK_RMENU},
    {"RSHIFT",VK_RSHIFT},
    {"RWIN",VK_RWIN},
    {"SCROLL",VK_SCROLL},
    {"SLEEP",VK_SLEEP},
    {"SNAPSHOT",VK_SNAPSHOT},
    {"UP",VK_UP},
    {"VOLUME_DOWN",VK_VOLUME_DOWN},
    {"VOLUME_MUTE",VK_VOLUME_MUTE},
    {"VOLUME_UP",VK_VOLUME_UP},
    {"XBUTTON1",VK_XBUTTON1},
    {"XBUTTON2", VK_XBUTTON2},
    { "", -1 },
};

int key_find(const char *key)
{
    int i;
    for(i=0; key_table[i].key[0]; i++)
    {
        if(0 == strcmp(key_table[i].key,key))
        {
            return  key_table[i].value;
        }

    }


    return -1;
}

char *key_find_by_value(int val)
{
    int i;
    val &= 0xFF;
    for(i=0; key_table[i].key[0]; i++)
    {
        if(val == key_table[i].value)
        {
            return  key_table[i].key;
        }

    }
    return "UNKNOWN";
}




int is_shift(int key)
{
    switch(key)
    {
    case VK_LCONTROL:
    case VK_RCONTROL:
    case VK_LSHIFT:
    case VK_RSHIFT:
    case VK_APPS:
    case VK_LMENU:
    case VK_RMENU:
    case VK_MENU:
    case VK_LWIN:
    case VK_RWIN:
        return 1;
    }
    return 0;
}

#define PRESSING_TIME 40

void press_key(INPUT *ip, int key)
{
    ip->ki.wVk = key; // virtual-key code
    ip->ki.dwFlags = 0; // 0 for key press
    SendInput(1, ip, sizeof(INPUT));
}

void unpress_key(INPUT *ip, int key)
{
    ip->ki.wVk = key; // virtual-key code
    ip->ki.dwFlags = KEYEVENTF_KEYUP; // KEYEVENTF_KEYUP for key release
    SendInput(1, ip, sizeof(INPUT));
}


void send_input(HWND hwnd, char *keys)
{
    int shift,ch;
    INPUT ip;
    ip.type = INPUT_KEYBOARD;
    ip.ki.wScan = 0;
    ip.ki.time = 0;
    ip.ki.dwExtraInfo = 0;
    //SetForegroundWindow(hwnd);
    shift = 0;
    while((ch = (*keys++) & 0xFF )    )
    {
        if(is_shift(ch))
        {
            shift = ch;
            press_key(&ip,shift);
            Sleep(delay_after_shift);
        }
        else
        {
            press_key(&ip,ch);
            Sleep(pressing_duration);
            unpress_key(&ip,ch);
            if(shift)
            {
                Sleep(delay_after_shift);
                unpress_key(&ip,shift);
                shift = 0;
            }
        }

    }
}

void send_input_hold(HWND hwnd, char *keys)
{
    int shift,ch;
    INPUT ip;
    ip.type = INPUT_KEYBOARD;
    ip.ki.wScan = 0;
    ip.ki.time = 0;
    ip.ki.dwExtraInfo = 0;
    //SetForegroundWindow(hwnd);
    shift = 0;
    while((ch = (*keys++) & 0xFF )    )
    {
        shift = ch;
        press_key(&ip,shift);
        Sleep(delay_after_shift);
    }
}


void send_input_release(HWND hwnd, char *keys)
{
    int shift,ch;
    INPUT ip;
    ip.type = INPUT_KEYBOARD;
    ip.ki.wScan = 0;
    ip.ki.time = 0;
    ip.ki.dwExtraInfo = 0;
    //SetForegroundWindow(hwnd);
    shift = 0;
    while((ch = (*keys++) & 0xFF )    )
    {
        shift = ch;
        unpress_key(&ip,shift);

    }
}


void send_msg_macros(SESSION_WORK_SPACE *sws, int ev_type, int ev_code, char *msg)
{
    int i;
    char comma[10] = {' ','-','-','>',' ',0};
    if(*msg)
    {
        printf(" on/off macros  ");
        for(i=0; msg[i]; i++)
        {
            printf("%s%s",comma,key_find_by_value(msg[i]));
            comma[0] = ',';
            comma[1] = ' ';
            comma[2] = 0;
        }
        send_input(sws->hwnd, msg);
    }
}

void send_hold_macros(SESSION_WORK_SPACE *sws, int ev_type, int ev_code, char *msg)
{
    int i;
    char comma[10] = {' ','-','-','>',' ',0};
    if(*msg)
    {
        printf(" hold macros    ");
        for(i=0; msg[i]; i++)
        {
            printf("%s%s",comma,key_find_by_value(msg[i]));
            comma[0] = ',';
            comma[1] = ' ';
            comma[2] = 0;
        }
        send_input_hold(sws->hwnd, msg);
    }
}

void send_release_macros(SESSION_WORK_SPACE *sws, int ev_type, int ev_code, char *msg)
{
    int i;
    char comma[10] = {' ','-','-','>',' ',0};
    if(*msg)
    {
        printf(" release macros ");

        for(i=0; msg[i]; i++)
        {
            printf("%s%s",comma,key_find_by_value(msg[i]));
            comma[0] = ',';
            comma[1] = ' ';
            comma[2] = 0;
        }
        send_input_release(sws->hwnd, msg);
    }
}




int on_input_event(SESSION_WORK_SPACE *sws, int event_type, int port, int event_id)
{
    char *hold, *on, *off, *release;


    if( key_mutex == INVALID_HANDLE_VALUE )
    {
       key_mutex    = CreateMutex(NULL,FALSE,NULL);
    }

    while(WAIT_OBJECT_0 != WaitForSingleObject(key_mutex,INFINITE));

    switch(event_type)
    {
    case BUTTON_PRESSED:

        hold = key_config[event_id + 256 * port].hold;
        on = key_config[event_id + 256 * port].on;

        if(on[0])
            send_msg_macros(sws, event_type, event_id, on);
        else if(hold[0])
        {
            send_hold_macros(sws, event_type, event_id, hold);
        }

        break;

    case BUTTON_RELEASED:

        release = key_config[event_id + 256 * port].release;
        off = key_config[event_id + 256 * port].off;

        if(off[0])
            send_msg_macros(sws, event_type, event_id, off);
        else if(release[0])
        {
            send_release_macros(sws, event_type, event_id, release);
        }


        break;

    case ROTARY_SWITCH_ON_POSITION:

        send_msg_macros(sws,
                        event_type,
                        event_id,
                        key_config[event_id + 256 * port].on);

        break;
    }


    ReleaseMutex(key_mutex);

    return 0;

}


